/*

*/
import scene.*;
import scene.primitives.*;
import pipeline.*;
import framebuffer.*;

import java.awt.Color;

/**
   This example shows that a triangle may be
   clipped into as many as ten new triangles.

   If you run this program you get four images of
   a red polygon that is the result of a triangle
   that has been clipped along three or four edges.
   To see how many triangles the polygon is made of
   (that is, how many triangles the original triangle
   got clipped into) you need to make a modification
   to the renderer's triangle clipping code.

   Open the file Clip_T.java in the pipeline package.
   At the end of line 391, add this code.
      t_new.setColorRandom();
   At the end of line 431, add this code.
      t1_new.setColorRandom();
   At the end of line 432, add this code.
      t2_new.setColorRandom();
   The new code gives the clipped triangles random
   colors so that we can see them in the final image.
   After changing Clip_t.java, recompile this file and
   run it (you may need to delete all the class files
   first).
*/
public class TriangleClippingDemo
{
   public static void main(String[] args)
   {
      // Create geometry for a Model.
      final Vertex[] v = new Vertex[6];

      v[0] = new Vertex( 0.0,  1.5, -1.0);
      v[1] = new Vertex( 1.5,  0.0, -1.0);
      v[2] = new Vertex(-0.5, -1.5, -1.0);

      v[3] = new Vertex( 0.0,  1.5, -1.0);
      v[4] = new Vertex( 1.5,  0.0, -1.0);
      v[5] = new Vertex(-2.5, -2.5, -1.0);

      Triangle t1 = new Triangle(v[0], v[1], v[2]);
      t1.setColor( Color.red );

      Triangle t2 = new Triangle(v[1], v[0], v[2]);
      t2.setColor( Color.red );

      Triangle t3 = new Triangle(v[3], v[4], v[5]);
      t3.setColor( Color.red );

      Triangle t4 = new Triangle(v[4], v[3], v[5]);
      t4.setColor( Color.red );


      // Create a Model object to hold the geometry.
      final Model model = new Model();
      model.doBackFaceCulling = false;

      // Create a Scene object to hold the model.
      Scene scene = new Scene();

      // Add the model to the Scene.
      scene.addModel(model);

      // Create a framebuffer to render our scene into.
      int vp_width  = 600;
      int vp_height = 600;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);

      // Render four times, once using each of the above four triangles.
      model.addPrimitive( t1 );
      fb.clearFB(java.awt.Color.lightGray);
      Pipeline.render(scene, fb);
      fb.dumpFB2File("TriangleClippingDemo_1.ppm");

      model.primitiveList.remove(0);
      model.addPrimitive( t2 );
      fb.clearFB(java.awt.Color.lightGray);
      Pipeline.render(scene, fb);
      fb.dumpFB2File("TriangleClippingDemo_2.ppm");

      model.primitiveList.remove(0);
      model.addPrimitive( t3 );
      fb.clearFB(java.awt.Color.lightGray);
      Pipeline.render(scene, fb);
      fb.dumpFB2File("TriangleClippingDemo_3.ppm");

      model.primitiveList.remove(0);
      model.addPrimitive( t4 );
      fb.clearFB(java.awt.Color.lightGray);
      Pipeline.render(scene, fb);
      fb.dumpFB2File("TriangleClippingDemo_4.ppm");
   }
}
