/*

*/
package scene.primitives;
import scene.*;

import java.util.LinkedList;
import java.util.ArrayList;
import java.util.List;

/**
   A LineSegment object has three Vertex objects that represent the
   corners of the triangle.
<p>
   A Triangle is the basic geometric primitive that we use to build
   "solid" three-dimensional objects (as opposed to wireframe 3D objects).
*/
public class Triangle extends Primitive implements Rasterizable
{
   /**
      Construct a default (empty) LineSegment.
   */
   public Triangle()
   {
      super();
   }


   /**
      Construct a Triangle from an array of vertices.

      @param vArray array of vertices to put into the new Triangle
   */
   public Triangle(Vertex[] vArray)
   {
      super(vArray);
   }


   /**
      Construct a Triangle from a List of vertices.

      @param vList list of vertices to put into the new Triangle
   */
   public Triangle(List<Vertex> vList)
   {
      super(vList);
   }


   /**
      Create a Triangle object with references to the three given Vertex objects.

      @param v0 1st vertex of the new Triangle
      @param v1 2nd vertex of the new Triangle
      @param v2 3rd vertex of the new Triangle
   */
   public Triangle(Vertex v0, Vertex v1, Vertex v2)
   {
      super();

      this.vertexList.add( v0 );
      this.vertexList.add( v1 );
      this.vertexList.add( v2 );
   }


   /**
      Create a Triangle object whose vertices are copies of the three vertices from the given Triangle object.

      Create a Triangle object with references to copies of the three
      Vertex objects in the given Triangle object.
      <p>
      Notice that this is a "deep copy" of the given Triangle object.

      @param ls Triangle to make a deep copy of
   */
   public Triangle(Triangle t) // a "copy constructor"
   {
      super();

      this.vertexList.add( new Vertex(t.vertexList.get(0)) );  // make copies of each Vertex!
      this.vertexList.add( new Vertex(t.vertexList.get(1)) );
      this.vertexList.add( new Vertex(t.vertexList.get(2)) );
   }


   /**
      Make a deep copy of this primitive.
   */
   public Primitive makeCopy()
   {
      return new Triangle(this);
   }


   /**
      Implement the assemble() method from the abstract base class.

      A Triangle should return a List<Primitive>
      object containing just this one Triangle object.
   */
   public List<Rasterizable> assemble()
   {
      List<Rasterizable> result = new ArrayList<Rasterizable>();
      result.add( this );
      return result;
   }


   /**
      For debugging.
   */
   public String toString()
   {
      String result = "Triangle:\n";
      result += this.vertexList.get(0).toString();
      result += this.vertexList.get(1).toString();
      result += this.vertexList.get(2).toString();
      return result;
   }
}